#!/usr/local/bin/perl
#
# read Adobe PDF Files (PDF-1.0 + PDF-1.1) and convert to PostScript Level 2
# Detlev Droege, <droege@informatik.uni-koblenz.de>, Sept. 1993
# Copyright (C) 1993-1995 Detlev Droege
#

# RCS: "$Id: pdf2ps.pl,v 3.5 95/09/06 14:28:44 droege Exp Locker: droege $";
# RCS: "$Source: /usr/bs1/droege/pdf/pdf2ps/RCS/pdf2ps.pl,v $";

$Version = "2.17";
$RCS_Revision = '$Revision: 3.5 $';
($dummy, $Rev, $dummy) = split (/ /, $RCS_Revision);

require "ctime.pl";
require "getopts.pl";

$logging = 0;
$noType3 = 0;
$opt_h = $opt_v = $opt_3 = $opt_b = 0;
$streamOutMode = 0;		# will be set to 1 or 2 during binary output unless -b

$UsageMsg = "usage: $0 [-abxhtv3] file.pdf
  -a	produce ASCII85 encoded data (default, SLOW, smallest pure 7-Bit files)
  -x	produce HEX encoded data     (fast, rather large pure 7-Bit files)
  -b	produce possibly binary data (fastest, might mess up data transfer)

  -h	print this message and exit
  -t	don't generate seperate PS file with thumbnails
  -v	be a bit more verbose
  -3	fake all Type3 fonts as Times-Roman";

&Getopts('abxhtv3');
die "This is $0, Version $Version (Rev $Rev)\n$UsageMsg\n" if $opt_h;
($pdffile = $ARGV[0]) || die "$UsageMsg\n";

$noType3 = $opt_3;
$verbose = $opt_v;
$NoThumbs = $opt_t;
$avoidBinary = (! $opt_b);
$opt_a = 1 unless (($opt_b) || ($opt_x));

if ($opt_x && (! $opt_a)) { # use HEX as default
	$avoidBinary = 1;
	$defaultAsciiMode = 1;
	$defaultAsciiFilter = "/ASCIIHexDecode"; # must match !!
	$defaultAsciiProcSuffix = "_Hex"; # must match !!
	warn "Option -x overrides -b\n" if ($opt_b);
}
if ($opt_a) {	# use ASCII85 output
	$avoidBinary = 1;
	$defaultAsciiMode = 2;
	$defaultAsciiFilter = "/ASCII85Decode"; # must match !!
	$defaultAsciiProcSuffix = "_A85"; # must match !!
	warn "Option -a overrides -b\n" if ($opt_b);
	warn "Option -a overrides -x\n" if ($opt_x);
};

$pdffile =~ m:(.*)\.[Pp][Dd][Ff]$:;
$basename = $1;
$basename = "x" if ($basename eq "");
# print "Basename: $basename\n";

open (PDF, "$pdffile") || die "Can't open file '$pdffile'\n";

# let's try to figure out the line ending char, \012 or \015, to set $/
while (1) {
	$c = getc PDF;
	last if ($c eq "\012");
	last if ($c eq "\015");
};
$/ = $c;
$c = getc PDF;
$/ = $c if (($/ eq "\015") && ($c eq "\012"));	# handle DOS's  CR LF

if ($logging) {
	print "In this file, lines seems to be terminated by ";
	print "Carriage Return (CR)\n" if ($/ eq "\015");
	print "Line Feed (LF)\n" if ($/ eq "\012");
}
seek (PDF, 0, 0);
# done. now let's start.

$_ = <PDF>;
/^%PDF-/ || die "File '$pdffile' is not a PDF file\n";
if (/^%PDF-(\d+).(\d+)/) {
	chop;
	$pdf_version = $1 * 1000 + $2;
	if ($pdf_version == 1000) {
		$pdf10 = 1;
	}
	elsif ($pdf_version == 1001) {
		$pdf11 = 1;
	}
	else {
    	warn "'$pdffile': never tried $_ files - let's see ...\n";
	}
}

sub put_proc_set
{
	if (open (HDR, "header_for_pdf")) {
		while (<HDR>) {
			print PS;
		};
		return;
	};
		# if header_for_pdf not present, use this stuff:
	print PS <<'EndOfProcSet';

% PostScript definitions for PDF interpretation
% Copyright 1994-1995 Detlev Droege.
currentfile /ASCII85Decode filter /LZWDecode filter cvx exec
J07)[i!\%T!!!$b"FpQ>%g`4F^]s"K$ip<WSRDL(Zsklu!It/8"oLmkM3.cJ!uOFK
rXOq2L2ja0)up3_63&aH,fVJ:)UG/l3<0$n!&?q'!DicH@4d_cBV<-A@[$j-!"o@e
KT;i.OFk7Ng:2M1p_DCkP%bA%K+A*8&ASK8StSND<<f]'#XAA9MU?qT]ANb/&@_]O
J-A2umf0"O0[Q=<495'p<%.1Yc%#gKDeEV]?if:.P-8'i!>#>l7@sPi"BMQ$ZiKur
ahRr8!2)]deQc4Fr?L5d^nT)?](m;?O/TJ6CeI>VmL]fHmW)BdLJCPr#0d,JJ-1:B
."lZHq&'V^X65K=]6O_srZ@^<J/)8aaWl&o2Q^']9XWe'HN0_a8t_Y$%"#jc^r39[
KEVoRU=DAZCL.B*T^*Yd4`:re;*)fe$nOcm"&t3k&Lp[(75.B/cQm%*?BBmRN%Uj"
"KYk!Kb;fKcQ`SBI\t1`RR2rHE]>d!^aR#$,D$P]<e$3o3"[Z2L\XdlJ.gq!J!1Ib
N`ZCD7)hGb^];2-`DYq\(G6iKi>G68q#reZ#(SWi,UjTDf<-Aes#A#K!=qKs!X0)s
9'#W)*)Z7]`_OV6Ke"=gBDEu$KbW=,T9SkeYu+:C<*YNU!7_(6OKEhr+B?o[3_[tE
XW]<\oZ_MO;l!AtOb=JJ31lG$(]XhDeJ^,7fg#_14HHd.n-mk3)$]4'#=Yn%-f`EU
dME\tG;58(;,MC:#tU)H[nS"8G$*uH")\PaV4<$tdQ#'k5UdQLf;E)t!7`?)lX&-\
!6)5D%sT;_LKk-'"aD="$"8]roEYFZ[HW3mNa8V$rJWPK7S.oV"1qn1'`brh3l^I9
AW!`q6:*eHq!(a;[HLG\T&cOEq33uh6ogbI3N8^KkK+Rtf7Ht>Tq[OWH4d4>KEb8Q
'FJEeSW/kAZ*@r`f[eZO_$8;]NAf`l$=VsYJcH?$V8/$aLdL*`NZ6DWnK]LI9XqH]
#6^[4J-68f1TQGT!-)o;`]uT>p],P'0:s/X[Ocn)cTFjWkjPk51d[q8i);[pEW(A(
nKu:p:lKH9>nV;Lkt/_:n7jFBEHeN.@Gl<X&--QkY9;fbElk6[Ech[1H/0_t+h?lg
*[e1]J-MXVk'QD*U$`!kl9^SkT2T,JIE;bmpH0WYiW'76EaV)=8I0kDMhj<@i6hT*
!*8uR(b'mQ=:I>97PnD]oJn#oIP1h7qG;g/IEo7p2oD_D843,o7X<H/GB.>eWsA\9
?5,O^<;2@Cg4`=c(;;!C"<R@`)<@0L*^F-().:^EKiA$\#=-d(!6^1/N*&JkMDAWY
+VPGJQ4/#4hBJ\h,\%T9P,P\%5QJ]XA0<j#pf&qW**u?-=K3<m*du.L1-VB/qM>9$
GE?h;&kCIOT/7U:j+0E$Vq".b"0>*gGZ#bsTccK!r^Dck+$O];oq48q:W!<n!!a7=
T.m5d%hU$$P$Er;Oq;<MUn&g[I)j>qINHe>/sOPt!""K]:ZZ,NPTTs58jWJ.jZmLF
jC2AWl!-[nWK3DZCtQrds"R+C`uj:>=NQCVB`LQHRKR)$&i$RCVHK]6#-q#s"M5!%
Ph.XRP#]e<`kQ+bl44Q6>YaIf!!U=!P=1?d:.T[t%'fC2Ze3BY$2#ah8FT"\[%`m!
<cpsgLNk.jYmLa\!@FcQ*/h[$OgtokM=IA[1AEe*)F.7N'YXRR<Toemk6+F6\sUNo
dL<\sDQ'QGTnZ+JM$?dI`.ALG"H5TEU45GiC%Vh^\<dAF6P<u#SN^MqnK'q+"@;CC
4G(G(cWUn18&[qB!@3YdU6m._L]O?)ri$KBblg?&opKelY099!</D"[WtXh"#`@q(
cOi[Dk%F`s3^-D4d2O;5@$_.FSf;'oOWdFu/_6>j>KhM81#XaK=Hae29RTg6[WTVo
9,AhprJiVkij+(7BE/%oQ7<UASLE1U?)gQ!!!9a$P(M[)-k/6^`?*kQqP[8+[<nSN
`+TtIFB&H4_5QGki^n4."3ja)P4r#';?C-:%!b!_kSO?J$;X(X*P6+QG3N$gmX8t5
T"-13XO)Q8'J=Kd+a5o4<.d>Z*C^%@Zq-hI7k@</OY3c>JdU2"j]78hq4lpnr`qN.
:PlLh5'CW8h?,'J0\JY*=GGAT8b$um<T--=7#PKD;/oC%b99,[gXlmVN%o,_`U`m:
;*")/>Oa-M(t?X\49oXtaQ]&>8<D5uI68Mr\XiDp=rnm+2J)OEGr"Z+j[VpDUqGQN
fWo26\\;SEWC<X15&+"Ib*R0`o?Au/W#5>oP]GbncK)StQ5GOIA`dKqKT9C$N(oBE
S?U8eE=i`['qD?8.68*&7?@8U!&UR7.&H]68TI-EF;@c;\q<@a:n;;K$1g5KK.k^5
&:q#K>:%)[lXXcAXnB&Vf2SAI>"Ur\md@isRce,Fc%3"j(.Y1k*.R;AMm.ARaP&]h
APrtu[5IBqg,gZdP3^&6pA$>jjutgCq$2>G_j)>:P0i(q+sEfUof#;*$oT;+6AL<\
^X=*D&ejE?&"P$##!BG^VS.1tp:%F&1nPC8#qrOB11L/!GVS-CTK'WkF(XL"?t^g-
!`'Z"_68U&M(W%"I(g5hq!oJA#VsZMfK&_S!"9oYZ&%)/]=]cGT2MWY\B3+],ZqR;
%QT)[?R=bZX.baV87TP'odileI.VG6`-%1k$*s>$EaGbUft*%)HV,9.%YLm^T%KgV
];M8`(<Noc8*Q356h,%;dS!YOj0J-e'E57fa\R=1I5Hn+fakSarr=;j`*@Fns53VL
+QaM$n!=X4j1L#(/4-hiGCq4(q7a*Y!Oi=EL5BdRJq"1d7qgK3!Tu*":GMcnepBA?
!'2d(d_8)e&V,UaI"ch>"]r>#;&O`EBE7-X;`k]\=)PRXbe4VpqX-3Sf(ssj0&;kt
T0B%'HTo*_h?K!h/@X>oYAGT$kk_c`k0hP,mkOiEMfR\p%aV<@k(gUiMKA(^!aJu+
k2h6<n\BhX/#WFYpc`Wo.M(l#n*i'nT`NLI1%5Z-n,OitKWpOR-.!Po@,CqJFpN#D
cNZK;YY+b1$KSEfi+(m\Xdgt>829UA/PoX/0Ok:UE^il<#7tRUJjqKV^-6#S1#Pm8
Nc`kAJj*sM#_VFKUZMu'da=aL5-;P\m9i+n5X'(_5%s#d:ce'k4B&Ja`CmfN,$bRc
>mmUjWj!tkOA%%\-OS$f3tSda6gG-?4?_#5%O*V;E6Qo&m0tc=1%=f\0HUc9<$1'u
d@i*TX?;U6QOnck%>''@iroiDY4KM9&Z!#NlSf(OlO?11'#H@:&JEZ73%*ZC5+o9b
@5:$?1d7S8X^NVAYJ1HtlKf^"O<QCY9uOE"ga13G-g:Qs&Qo?JIhW2#.)E%h@!N&d
'cGqSm=I8'@7Uk.q?QPq)aJiud%hfsV&GK.)f_u"68oi<[iBC]*!>L/n_"g)2]?\0
$"/i7+S?\*2]hM[-,Kf-;0kF!7'1/KOR@SJ6gBa&&-)]Rr2UD)U42l(iW@J)-h3(j
UW<Ja.O,QYnfn8A1,gGqa9==cEMLSf_H>Iuf+`ukgC91rWXqG3O?)<*+cCFEk"msE
0W$%!6kUXf/h25K-O\t]QEk`_mqSbN!@!=I-tBFX_D*JjE^?Z&'MsJX%)46M4"C]N
&n6=rC!n`oi3.]O-AoJeD@lKDHMN^?,>gf'OX*,r^m`]iYCTP%cVV7;k79?ANT9=@
_D6:b!@bqKac_>cn*0lc;([tA`3T=rj!6pII2Ls%O0ONb,4[YPEb!G,-.-c2DfZ*2
5QD<J1eb@JF%GpR]enci%I/[,ULt&jpn7aY!u#/A@ZL"saYln02,;a\Nu9fM4S)?U
2pXW)!\kRVO<="[ae/VJ;#:*"8/1ZilQS$X$4pf8;)[E_%'*S(62m]q:]M0-qG+=>
;kZ%b0FTBe-Xbjhl\$Gqdb,$Z4`f<W!73RgoF)Dk_Tppae'U4N0GPhO5VpXR!\b=`
lD?9M-fFndZ3mAN.mV#7#QsBY6Q\UV<-2(_38ujnC+'&^0/Z=63PZpu5lS^;%(5Hd
7;UJ;FBG^h!"lHLT&AHRK5;)UCGhcQ7gYOb1h-5g;b(=V,qSs(e93["3AY%C4<]PO
@R^QDHo.'n7C7\>PcmOeOoTTO.,jbf@]0AcHorMh8p%GcJ6,F[91Qi5*;#3b@BWA@
/P1P%(=\uW!+S^PWh/kVpgM'1@MB+-LlgRe%DeeIkmeEO^(I\N'bK.<<.$;A_+2A\
'f?1?@9Cr#lV,,*8V2sC1s?prhDBa180[%02(3.fQolMK8V.Wp<.#s"Bi<q%(?<=[
!B3MXC`*rD/3q<KJuB?jBE/'XdNGqhJntd:<WE,Z2*lc2^k5=r[7=]E%<RdWPp5r;
>]J8\,[lg*_oNpaZ3bZV*n5cm5f#s"iX.QG"a-sR[AX'*F^7u:7q+djFI4n9iCb:4
>2W%[+>m?\CKht.80c1\Fn,M5KhT>M<+9cP+o+jliZ$CQ#.:lD'g)Hsk=i^g9-LQF
huFVdZ546LQ^!(F6"`_47MBp[!!-.o+[8UC94S&PA>'f4+TG(X$Y08%A>%OF,!lIA
&e).Q$T,-sp-8r6Z3U(q#dglYFfl@abQ`0p#<"gN(.n]$Z39j/"B:a@+k9F(50DAD
7$J#Y'Mc96!6a:6>@7b/Q,?6Pbsh=!q)/TU2CV)g5=P9,3`\UYTm^Hp:&l&-\JAI_
<afU+D.cmhB9ms=2MjlrO[\u'Dq:0K2P!;1R76hO7J<^!f!+s.[70_=7J09b[0hLn
)+ug_/g:*b[P`4C5l!8aq)%(Z[9mLQ+Xsee>P)92CUt94^Ju(qGb[AH+`17&!EOE'
H;b,qeTEQiUNtlI;So:2'npq-;e;2#C0.PM2dd0j@8hBW1mDp*6>\E5%N+6E4<VU5
_I*_:>=Z/A!#5Z+oqOt6G>-QlGuUSb<:`QNdm`o\HM#-([)lN/\@5GP7J<F%<l&+,
s)>f\7q9Y;p^dE>Sg^u3:>on^iFQ>[LP(*X=#`c@Q`inP\f@$X8J_8.f)8l2?WqB3
?'J]$P[@`4rEaE*"uY.BecC)>Y@"YC!;7REef)/om*['1:%[1KeO=q\\RJ(H:#l^#
o\W_lhC=)%@DQf;^]5-*;NU&h+UI[gTiiZu2El&.Hf>_P37K,hdP9\.5-d5nJea;]
So\:A1%!UK=PljD:Q/f%kA(jlP8umITo_&I4Fp(a0fP4=/Wk\K+D`\d(tQN!Oa,da
'A4KMmn&YJ'pJG[5&dfm!5L+;F<k7fPtV!A\M$7Z..4Cj,7p:idZ]WWm5#U,POBZS
YiCn7-IWhsNh`oqGk34*:b%"r"o/$]!!/)r%5E,/R$jk9m!Bp$l3pN[/jnqR.4Q#k
j-#QARZd;e6;T#F^]iD_h74>j3IM@i+$efb.RG;I=XtVXs-$ns^+B0nffn7^1HubI
2aWN,)@MhiTHlP:4MjDr)<6i./#!ptD]&hmY4d5=+/1M6OuX8Pkpqt8+-a:eS@M0S
\PAQDiLmM`*#]JTbRp-YAY'O_g:7bYH0ca`7+j:CcNt*nE\3WG3"UBqr-C:b:-Ku>
#+D8RW0%--dcMdr2-4^,-.r5?3S^WXae0%*P<qSD)3cmun1j8,,2/NMms=N`IAlcW
.OPOsg+/nWZ9E>bXnq.C1TVnk4:NFg%fkAn+PX)MkR=-0%%uYkJFX0Mapl/-%3WY#
J;OfO`XVe`%D^6p^letm((=T0%H-"OJ=WnEbHdTW%Y2P-+Q]rFh@09D%%u/\!:-rF
:_'/P%=l;*]1`XF)@`0P%U/W9i5cAckR8TR%!ra"JA_p_apu50%D.J)4$F)rfF3;Q
$t.Kn5f^<`bRK*F%fiU<i8b;.QjUI-],S/iS%Z>$E"&D3]GnctTNhr\Kt;fY$2_oX
+PXMT\dA8aY[E[.gR&:/)mI-#%Ud]p@$A>%cb#j6^S&^T]E,kH#J"(J%Uaq*^l-$5
o=Y-V_cuaD)\oU8cjepu%Ud2ZJDq'NKss=c%=lA*+Qa*F;oO+nZJ8a*!6`;'e$>I&
Zht(;JB/-Ac`j)'%%LP6@*B'&c47I_`h:TdRsYY&cjlS4ah*&3JBA;T9Fd`W\Q^<i
+O.P\+3.dE$TofP]FGH7cbd:4%`$Q)S7f'RUWRuD%(j'n>jKJdTugR(`ENSn>jnn)
gW1N"%3$,1S4C8CM@&5*\m)*=4WCt?ap:*R%Ue8*JBQ6RfF8t^_Ruqm!8G"gkK9<_
e/O+g415UqHk!mE[<il%+PFJ5"qL^S[bV,si9h'0;o$'cZeL,a+PmTTD:"rn[![eB
^ur<m2@\K$%"QU5](rP!RLZ"^Z!cK!S4g(/h8GtYfN_en!.kKhf>.;(%\(AIS"'sR
f<SfB%m<Fn5gR7DjgSKO\_A"@]e-W3KtJPd%04<V]W_m9A'B@LYa^oe>_U7LleGn1
%_Os<!:.KElbV>0dG%1MJD]>_]=!"'\U`SRgS<2%h@2RR_Rgi"qshT%Q/eO[%6PZ?
?:i#(&%]JP_-lfkgYN@O*Orgl[t)B()h&2Y,J&ECiL>\d]JmI9pV$QU]/p3(^).1U
(^]&S_>G8YH]>L]C^<1)%,ccsS0Q=lS\'F:`*hX8_"6[eapO`bbe+o)^&cWpXq)Y5
m(U[$r):!Cj:6N?%38gbgKX,.&$XemiOj<`?LQ?-%C"W3Yi,!&+RAOTj:*%Ad(nJJ
+PmQS(WG=offN';^o+R:TEW.$]7A`0Rf3bhJd'rC\q&iJ].(H8c1F/-^&Z4B^uPiN
0=\f'nbe=>>qcGtkM4rM%%r=ai7J4gPII0Fq)r="J>MC9?2<EHk<R)uJ=G78kR%<G
#iCF$5g0%TNNp):#hu-^)c@)Pjp6HVr>m^(@"m+'a9^_Wo"-uii.MA$h@C^sq]:D#
63(^^&WPr5Tht<!15h+nj1,FpF?CQS-Lti[k[fUe+OQ"md&(M?),"H@k^_`*2]3%%
+T\t-/.`<s_?#1;k%-RZ"<nJPOJki&)N'8)U"ndP3>tW3GbcaND(qC%UsfXoej:br
a@;Aebb6U3S'@i:/?%c".G2n"o8u`)@540@L/T)3kYSkHjb3dCaP(\[)GQ?XOW_e\
1R\f,P%c6P/[HB)0p9)CaE06O8:h5^(rK$-L^!Ob^pqC*+rX-f6N*;6DD7Pm:m3Hk
l$'S+NcuG64Jl%&PDJ>k)*E^D*=)p_(bhIFH(>3eL(dsFmp7hb1tUL]"@5]mB*H:k
-6$$g(6OP'JHqYP/b#7BA3k.Od?QDMc0&#gjroOi8>n"$K>Y8UL$Tc<QOVsl`r?IB
,Nej&.[mB[WFNjtWZ_XI4>MY$+Cau'K+Q*JMZOHo-jjr2a;>2W9/n($+n0mANIi"9
/;\<SE[_:R&;DW(+unX;O!5F$a<cJW,eB#\@*9[K;QQWBaRW$0!(W1D]H.<0OQnb2
M,Kj#5m!.8-4</Q&N7nn(o[WDPFM\!#V0.l*!WW4K0IUX&<assla$"&M%DZ'.LABP
:+?E6T\DL7K@$PWMM%"h1C0_JKIR>f:rF(W_,=2SN^GeG2N=J6^+kX8GimY<j*DE)
MJ:H&3(*e7b_!dN.*e\k6`#'8=:C'f^FG+5r]42h38coWj.P8\ejCAU,S)E=WG>]F
8sZ@)_YDnANoKn70s]hYJi=0L8;]>ZMI=bS?5Jf+1CVm[i\]PF29g5K=(86cMG#1b
H="a3&VV]%Fp3\1GaQs/gt,rKAIb%\ac$Y\jb&.j2dP^Rg59)MSV7+^;@j$UaDr/P
\WChMKF=aB2@Q@hF]hgV-pSCP=n'mc+KegW(Bu&&L_:"=,F1.pUC(*`A9u>T1(;!2
J#6qr#&kB5!Pi"Z#(Zbb)$`H^g)03E-A&6#!\!Ot"9N*R25RH^ju;L-+p%~>

PDFView begin /'  { /ty ty .textLead sub def tx ty moveto Tj } bd end

EndOfProcSet
};

sub find_startxref
{
	seek (PDF, -50, 2);
	local($found_startxref) = 0;
	$startxref = 0;
	while (<PDF>) {
		if (/^startxref/) {
			$found_startxref = 1;
			next;
		}
		if (/[\012\015]+startxref[\012\015]+(\d+)[ \012\015]+/) {
			$startxref = $1;
			$found_startxref = 1;
			last;
		}
		next if (/^[ \015]*$/);
		/^(\d+)/;
		$startxref = $1 if $found_startxref;
		last if ($startxref > 0);
		$found_startxref = 0;
	}
	warn "Could not locate 'startxref' entry - truncated file ?\n"
		unless $found_startxref;
};

sub load_xref
{
	local ($start) = @_;
	local ($first, $num);

	seek (PDF, $start, 0);
	$_ = <PDF>;
	chop;
	/^xref/ || die "XREF not found at pos. $start as expected\n";
	print "loading Xref at pos. $start\n" if $verbose;
	while (<PDF>) {
		# print ">> $_ <<\n";
		next if (/^[ \015]*$/);
		last if (/^trailer/);
		($first, $num) = split;
		for ($i = $first; $num; $i++, $num--) {
			$_ = <PDF>;
			$_ = <PDF> if (/^[ \015]*$/);
			chop;
			/^(\d+) (\d+) ([nf])/;
			if ($Xpos[$i]) {
				# print " ignore Xref '$_' for obj $i (use $Xpos[$i] $Xver[$i] $Xuse[$i])\n";
				next;		# don't overwrite existing entries
			}
			$Xpos[$i] = $1 + 0;
			$Xver[$i] = $2;
			$Xuse[$i] = $3;
		};
	};
};

sub token
{
	local ($token);	# $tbuf sollte 'static' sein ... :-(
	local ($tktyp, $multi);
	
	$_ = $tbuf;
	$multi = 0;
	do {
		$_ = <PDF> if (/^$/);	# next line if empty
		s/^[\s\015]*//;			# strip leading blanks
		s/^%.*$//;				# strip comments
	} while (/^$/);
	if (m:^(/[\002a-zA-Z0-9\._][^\s\012\015%()<>[\]{}/]*):) {	# /Identifier
		$token = $1;
		$tktyp = "Id";
	}
	elsif (m:^(\d+\s+\d+\s+R)[\s\012\015%()<>\[\]{}/]:) {
		$token = $1;	# Reference: 	num num R
		$tktyp = "Ref";
	}
	elsif (/^([\[\]])/) {	# '[' or ']'
		$token = $1;
		$tktyp = "Array begin/end";
	}
	elsif (/^(<<)/) {	# dict start '<<'
		$token = $1;
		$tktyp = "Dict start";
	}
	elsif (/^(>>)/) {	# dict end '>>'
		$token = $1;
		$tktyp = "Dict end";
	}
	elsif (/^(true)/) {	# Constant:	true
		$token = $1;
		$tktyp = "const true";
	}
	elsif (/^(false)/) {	# Constant:	false
		$token = $1;
		$tktyp = "const false";
	}
	elsif (/^(null)/) {		# Constant:	null
		$token = $1;
		$tktyp = "const null";
	}
	elsif (/^([+-]?[0-9]*\.?[0-9]+)/) {	# Number	# OK ???
		$token = $1;
		$tktyp = "Number";
	}
	elsif (/^([\w]+)/) {	# Name (\w == alphanumeric + '_')
		$token = $1;
		$tktyp = "Name";
	}
	elsif (/^(\([^\)]*\))/) {	#	String:		(abc)
			# (string) - multiline stings see below !!!
		$token = $1;
		$tktyp = "String";
	}
	elsif (/^(\([^\)]*$)/) {	#	String:		multiline (abc\^Mdef)
		$token = $1;
		$tktyp = "String(multiline)";
		$multi = 1;
		do {
			$token =~ s/\\[\012\015]+//;
			$_ = <PDF>;
			if (/^([^\)]*\))/) {
				$token .= $1;
				last;
			};
			$token .= $_;
		} while (1);
		print STDERR "ML: '$token'\n";
	}
	elsif (/^(<[0-9a-fA-F]+>)/) {	# HEX-String:	<0fab12>
			# HEX (string) - can't yet handle multiline hex stings !!! ???
		$token = "$1";
		$tktyp = "Hex-String";
	}
	else {
		$tpos = tell(PDF);
		open (ERR, ">/tmp/PDF_ERR");
		print ERR "Unrecognized token: '$_' (near PDF pos $tpos)\n";
		die "Unrecognized token: '$_' (near PDF pos $tpos)\n ";
		$tktyp = "Unknown";
		return 0;
	};
	$tpos = tell(PDF) - length;
	if ($multi) {
		$_ = substr ($_, length ($1));
	}
	else {
		$_ = substr ($_, length ($token));
	}
	$tbuf = $_;
	# warn "TOKEN[$ld_nest]: '$token'\n" if $Problem;
	die "Null token '$_' ($tpos)" if ("$token" eq "");
	return $token;
};

	# load_dir: aktuelle Zeile ist die, in der '<<' steht
sub load_dir
{
	local ($tok, $key, $val, %val, %pairs, $len);
	$ld_nest++;
	$tok = &token;
	($tok eq "<<") || die "dictionary start not recognized near pos ",
		tell(PDF), " \n ";
	# print LOG "load_dir: at pos ", tell(PDF), "\n" if $logging;
	while ($tok = &token) {
		last if ($tok eq ">>");
		$len = length;
		$key = $tok;
		$key =~ s:^/::; 	# strip leading slash
		$tok = &token;
		if ($tok eq "[") {
			$array_nest = 1;	# first '[' already scanned
			$val = $tok;
			do {
				$tok = &token;
				$val .= " ".$tok;
				$array_nest++ if ($tok eq "[");
				$array_nest-- if ($tok eq "]");
			} until (($array_nest == 0) && ($tok eq "]"));
			# print "ARRAY in DICT: /$key $val\n";
		}
		elsif ($tok eq "<<") {
			$val = $tpos;			# just store position to find it.
			$tbuf = "<< ".$tbuf;	# push "<<" back in for load_dir
			%val = &load_dir;
			$loaded_rnam {"DPos_$val"} = 1;
			eval "\%DPos_$val = \%val;";
		}
		else {
			$val = $tok;
		};
		$pairs{$key} = $val;
		# print "LD[$ld_nest]: /$key  := $val\n" if $Problem;
	};
	# print "Dir loaded[$ld_nest]: "; &dump_dir(%pairs);
	$ld_nest--;
	return %pairs;
};

# pos_to: position to given byte in PDF file and read current line
sub pos_to
{
	local($position) = @_;
	
	seek (PDF, $position + 0, 0);
	print LOG "pos: positioning to $position\n" if $logging;
	$tbuf = $_ = <PDF>;
};

sub load_dir_by_ref
{
	local ($ref) = @_;
	local (%dir,$rnam,$isref);
	
	if ($ref =~ /(\d+)\s+(\d+)\s+R/) {
		$rnam = "DRef_$1_$2_R";
		$isref = 1;
	}
	else {
		$rnam = "DPos_$ref";
		$isref = 0;
	};
	print LOG "load_dir_by_ref: ref '$ref' ($rnam)\n" if $logging;
	if ($loaded_rnam {$rnam}) {
		eval "return \%$rnam;";
	}
	else {
		if ($isref) {
			&pos_ref ($ref);
		}
		else {
			&pos_to ($ref);
		};
		%dir = &load_dir;
		$loaded_rnam {$rnam} = 1;
		$streamPos{$ref} = tell (PDF);	# might be a stream which starts _here_
		eval "\%$rnam = \%dir;";
		return %dir;
	};
};

# pos_ref: position to a reference (given as string "num ver R")
sub pos_ref
{
	local($par) = @_;
	local ($obj, $ver, $o, $v);
	
	($obj, $ver) = ($par =~ /(\d+) (\d+) R/);

	print LOG "pos_ref: ref = $par, " if $logging;
	&pos_to ($Xpos[$obj]);
	($o, $v) = /(\d+)\s+(\d+)\s+obj/;
	if (($obj != $o) || ($ver != $v)) {
		print "=='$_'==($o, $v)\n";
		die "Illegal Xref ($obj, $ver)\n ";
	};
	$tbuf = $_ = <PDF>;
};

# pos_dir_or_ref: position to reference or byte, depending on par format
sub pos_dir_or_ref
{
	local($par) = @_;
	
	if ($par =~ /\d+\s+\d+\s+R/) {
		&pos_ref ($par);
	}
	else {
		&pos_to ($par);
	};
};

sub load_pages
{
	local($ref) = @_;
	local(*dir, $kids, $typ);
	
	# printf "[$ref] ";
	# &pos_ref($ref); %dir = &load_dir;
	%dir = &load_dir_by_ref ($ref);
	# print "Ref '$ref':\n"; &dump_dir (%dir);
	if ( ! ($typ = $dir{"Type"})) {
		# circumvent bug - some /Pages dicts don't have /Type fields
		# (seen in file generated by Mac Net Distiller 1.0 from FrameMaker3.01)
		$typ = "/Pages" if ($dir{"Kids"});
		warn "/Type entry missing in Dict '$ref', guess '$typ'\n";
	}
	if ($typ =~ m"/Pages") {
	
		# printf "($ref) Pages:\n";
		# &dump_dir (%dir);
		
		$kids = $dir{"Kids"};
		$kids =~ s/\s*\[(.*)\]\s*/$1/;
		while ($kids =~ /(\d+ \d+ R)/) {
			&load_pages ($1);
			$kids =~ s/\s*\d+ \d+ R\s*(.*)/$1/;
		};
	}
	else {
		die "wrong type: '$typ' instead of '/Page' in object\n "
			if ($typ ne "/Page");
		$page_cnt++;
		$page[$page_cnt] = $ref;
		# printf "($ref) Page %d:\n", $page_cnt;
		# &dump_dir (%dir);
		
		return;
	};
};

sub dump_dir
{
	local(%dir) = @_;
	foreach $i (keys(%dir)) {
		printf "%12s = %12s\n", $i, $dir{$i};
	};
};

sub gen_page_resource
{
	local($pg) = @_;
	local(%page, %resources, %fonts, %xobj, %cs);
	local($fdir, $fnt, $key, $xdir, $cs);
	
	# &pos_ref ($page[$pg]); %page = &load_dir;
	%page = &load_dir_by_ref($page[$pg]);
	# print "Ref '$page[$pg]':\n"; &dump_dir (%page);
	
	$cnt_annots++ if ($page{"Annots"});
	print "Rotate ignored on Page $pg ($page{\"Rotate\"})\n"
		if ($page{"Rotate"});
	&make_thumb ($page{"Thumb"}, $pg) if ($page{"Thumb"});
	return unless $page{"Resources"};
	
	&pos_dir_or_ref ($page{"Resources"});
	%resources = &load_dir;
	# print "Resource (pg $pg):\n";
	# &dump_dir (%resources);
	
	if ($fdir = $resources{"Font"}) {
		local($fn, $fref, $refnam);
		# &pos_dir_or_ref ($fdir); %fonts = &load_dir;
		%fonts = &load_dir_by_ref ($fdir);
		# print "Resources/Font (pg $pg):\n";
		# &dump_dir (%fonts);
		
		while (($fn, $fref) = each %fonts) {
			$fref =~ m:(\d+)\s+(\d+)\s+R:;
			$refnam = "Fnt_$1_$2";
			&font_prolog ($refnam, $fref);
		};
	};
	if ($xdir = $resources{"XObject"}) {
		local($xn, $xref, $refnam);
		&pos_dir_or_ref ($xdir);
		%xobj = &load_dir;
		# print "Resources/XObject (pg $pg):\n";
		# &dump_dir (%xobj);
		
		while (($xn, $xref) = each %xobj) {
			$xref =~ m:(\d+)\s+(\d+)\s+R:;
			$refnam = "XObj_$1_$2";
			&xobj_prolog ($refnam, $xref);
			$XObjects++;
		};
	};
	if ($cs = $resources{"ColorSpace"}) {
		local($csn, $csref, $refnam);
		&pos_dir_or_ref ($cs);
		%cs = &load_dir;
		print "Page $pg: unhandled ColorSpace\n"; # unless ($cs_warn_cnt++);
		print PS "%%Page $pg: unhandled ColorSpace\n";
		if ($verbose) {
			print "Resources/ColorSpace (pg $pg):\n";
			&dump_dir (%cs);
		};
		
		while (($csn, $csref) = each %cs) {
			$csref =~ m:(\d+)\s+(\d+)\s+R:;
			$refnam = "ColorSpace_$1_$2";
			&cs_prolog ($refnam, $csref);
		};
		# ??? ColorSpace ??? noch nicht behandelt		
	};
	foreach $key (sort keys(%resources)) {
		next if ($key eq "Font");
		next if ($key eq "XObject");
		next if ($key eq "ProcSet"); # !!!!??? noch nicht richtig behandelt.
		next if ($key eq "ColorSpace");
		print "Page $pg: unhandled Resource '$key = $resources{$key}'\n";
	};
};

sub setup_page_output
{
	local(%page) = @_;
	local(%resources, %fonts, %xobj);
	local($fdir, $fnt, $key, $xdir);
	
	return unless $page{"Resources"};
	
	&pos_dir_or_ref ($page{"Resources"});
	%resources = &load_dir;
	
	if ($fdir = $resources{"Font"}) {
		local($fn, $fref, $refnam);
		# &pos_dir_or_ref ($fdir); %fonts = &load_dir;
		%fonts = &load_dir_by_ref ($fdir);
		
		while (($fn, $fref) = each %fonts) {
			$fref =~ m:(\d+)\s+(\d+)\s+R:;
			$refnam = "Fnt_$1_$2";
			print PS "  /$fn /$refnam\n";
		};
	};
	if ($xdir = $resources{"XObject"}) {
		local($xn, $xref, $refnam);
		&pos_dir_or_ref ($xdir);
		%xobj = &load_dir;
		
		while (($xn, $xref) = each %xobj) {
			$xref =~ m:(\d+)\s+(\d+)\s+R:;
			$refnam = "XObj_$1_$2";
			print PS "  /$xn /$refnam\n";
		};
	};
};

# copy /Indexed color space table from PDF to output file (THUMB or PS)
# &load_print_index_lookup($lookup, $name, THUMB);
sub load_print_index_lookup
{
	local($ref, $name, $OUTFILE) = @_;
	local($tmppos, %idir, $filter, $inxLlen, $l);
	
	$tmppos = tell(PDF);
	print $OUTFILE "% Indexed lookup table object $ref\n";
	# &pos_ref($ref); %idir = &load_dir;
	%idir = &load_dir_by_ref($ref);
	$l = $idir{"Length"};
	$inxLlen = &get_val($l);
	$filter = &gen_filterChain ($idir{"Filter"}, $idir{"DecodeParms"});
	if ($stream_is_binary && $avoidBinary) {
		$filter = "$defaultAsciiFilter filter $filter";
		$streamOutMode = $defaultAsciiMode;
	};
	print $OUTFILE " { /$name currentfile $filter\n ";	# defer execution
	print $OUTFILE "1024 string readstring pop def } exec\n";
	# print STDERR "CALL(2): copy_stream_len ($$inxLlen, \$OUTFILE)\n";
	# if load_dir_by_ref returned a "stored" dict we need to set the position:
	seek (PDF, $streamPos{$ref}, 0);
	
	&copy_stream_len ($inxLlen, $OUTFILE);
	seek (PDF, $tmppos, 0);
};

sub setup_thumb
{
	return if $setup_thumb_done;
	$setup_thumb_done = 1;
	
	open (THUMB, ">$basename.thumb.ps")
		|| die "Can't open/create $basename.thumb.ps\n ";
	print "Thumb File: $basename.thumb.ps\n";

	print THUMB "%!PS-Adobe-2.0\n%%Pages: (atend)\n";
	print THUMB "%%LanguageLevel: 2\n%%EndComments\n\n";

	$thumb_x = $thumb_x_off =  20;
	$thumb_y = $thumb_y_off= 680;
	$thumb_x_inc =  90;
	$thumb_y_inc =  130;
	$thumb_pg =  1;
	print THUMB "%%Page: $thumb_pg $thumb_pg\n\n";
	print THUMB " /Helvetica findfont 12 scalefont setfont .15 setlinewidth\n";
	print THUMB "180 820 moveto (Thumbnails for PDF-File '$pdffile') show\n";
};

sub make_thumb
{
	local ($ref, $pg) = @_;
	local (%thumb, $filter, $w, $h, $bps, $cs, $hcs, $len, $mypos);
	local ($tmppos, $dec, $base, $lookup, $lnam, $hival);
	local ($tlen);
	
	return if $NoThumbs;
	&setup_thumb;
	
	# &pos_ref ($ref); %thumb = &load_dir;
	%thumb = &load_dir_by_ref ($ref);
	$mypos = tell(PDF);
	
	$w = $thumb{"Width"};
	$h = $thumb{"Height"};
	$bps = $thumb{"BitsPerComponent"};
	$cs = $thumb{"ColorSpace"};
	$len = $thumb{"Length"};
	$tlen = &get_val ($len);
		# must handle /CalRGB... (PDF-1.1) ColorSpaces
	$hcs = &get_val ($cs);
	if ($hcs eq "[") {
		$hcs = &get_ref_array ($cs);
		$cs = "/Indexed";
	}
	else {
		$cs = $hcs;
	};
	$dec = $Decode{$cs} unless ($thumb{"Decode"});
	if ($hcs =~
	  m:\[\s*/Indexed\s*(/Device[A-Z]+)\s+(\d+)\s+(\d+\s+\d+\s+R)\s*\]:) {
		  	$base = $1;
			$hival = $2;
			$lookup = $3;
			$lnam = $lookup;
			$lnam =~ s: R::;
			$lnam =~ s: :_:g;
			$lnam = "inx_lookup_obj_$lnam";
			 # PROBLEM: this load should go into the prolog of
			 # the thumb-nail file. Would need an extra scan for that.
			 # So I delete the loaded entries mark after a showpage ... :-(
			&load_print_index_lookup($lookup, $lnam, THUMB)
				unless $loaded{"THUMB/$lnam"};
			$loaded{"THUMB/$lnam"} = 1;
			$hcs = "[/Indexed$base $hival $lnam]";
			$dec = "[0 $hival]" unless ($thumb{"Decode"});
	};

	print THUMB "% thumbnail for page $pg (object '$ref')\n";
	print THUMB "$thumb_x $thumb_y 15 add moveto\n";
	print THUMB " gsave currentpoint translate";
	printf THUMB " -1 -1 %d %d rectstroke\n", $w+2, $h+2;
	
	print THUMB " /.im_proc { currentfile ";
	if ($filter = $thumb{"Filter"}) {
		$filter =~ s/\s*\[(.*)\]\s*/$1/;
		while ($filter =~ m:(/[^\s]+):) {
			print THUMB "$1 filter ";
			$filter =~ s:/[^\s]+\s*(.*):$1:;
		};
	};
	print THUMB "} def\n";
	# print "THUMB: /ColorSpace $hcs\n";
	if ($hcs =~ m:^\[:) {
		#warn "Can't yet deal with /Indexed ColorSpace (Thumbnail page $pg)\n";
		# fake it as DeviceGray ... !!! ???
	};
	print THUMB " $hcs setcolorspace\n";
	print THUMB " << /ImageType 1 /Width $w /Height $h /ImageMatrix ";
	warn "Can't yet deal with /Decode entries (XObject page $pg)\n"
		if $xdir{"Decode"};
	print THUMB "[1 0 0 -1 0 $h]\n";
	print THUMB "   /DataSource .im_proc ";
	print THUMB "\n   /BitsPerComponent $bps";
	if ($thumb{"Decode"}) {
		warn "Can't yet deal with /Decode entries (Thumbnail page $pg)\n";
		print THUMB " /Decode $Decode{$cs}"; # fake, use standard ???
	}
	else {
		print THUMB " /Decode $dec";
	};
	print THUMB " >>\n image\n";
	seek (PDF, $mypos, 0);
	# print STDERR "CALL(3): copy_stream_len ($tlen, THUMB)\n";
	&copy_stream_len ($tlen, THUMB);
	print THUMB "\ngrestore %currentpoint $w $h rectstroke\n";
	print THUMB "$thumb_x 15 add $thumb_y moveto (Page $pg) show\n\n";
	
	$thumb_x += $thumb_x_inc;
	if ($thumb_x > 520) {
		$thumb_x = $thumb_x_off;
		$thumb_y -= $thumb_y_inc;
		if ($thumb_y < 0) {
			$thumb_pg++;
			$thumb_y = $thumb_y_off;
			print THUMB " showpage\n%%Page: $thumb_pg $thumb_pg\n\n";
			print THUMB
				" /Helvetica findfont 12 scalefont setfont .15 setlinewidth\n";
			foreach $key (keys %loaded) {
				delete $loaded{$key} if ($key =~ m:THUMB/inx_lookup:); #old
				delete $loaded{$key} if ($key =~ m:^Thumb:);	#new
			};
		};
	};
};

sub gen_page_output
{
	local($pg) = @_;
	local(%page, $contents, $c);
	
	# &pos_ref ($page[$pg]); %page = &load_dir;
	%page = &load_dir_by_ref ($page[$pg]);
	print PS "<<\n";
	&setup_page_output(%page);
	print PS " >> begin startPage\n";
	
	$contents = $page{"Contents"};
	if (! ($contents =~ m/^\[/)) {
		$c = &get_val ($contents);
		if ($c eq "[") {
			$contents = &get_ref_array ($contents);
		}
	}
	print STDERR "pg $pg: Contents Array: '$contents'\n"
		if (($contents =~ m/\[/) && $verbose);
	print LOG "gen_page_output: /Contents '$contents'\n" if $logging;
	$contents =~ s/\s*\[(.*)\]\s*/$1/;
	print PS "% Page $pg ";
	
	while ($contents =~ /(\d+ \d+ R)/) {
			&gen_stream ($1);
			$contents =~ s/\s*\d+ \d+ R\s*(.*)/$1/;
	};
	print PS " endPage end showpage\n";
};

sub gen_filterChain
{
	local ($filter, $DecodeParms) = @_;
	local (@dec, $f, $fil, $dp);
	local ($dpar, $d2, %dpdir);
	$stream_is_compressed = 0;
	$stream_is_binary = 1;

	return "" unless $filter;
	
	$fil = "";
	$dp = $DecodeParms;
	$filter =~ s/\s*\[\s*(.*)\s*\]\s*/$1/;
	$dp =~ s/\s*\[\s*(.*)\s*\]\s*/$1/ if ($dp);
	@dec = split (/\s+/, $filter);
	foreach $f (@dec) {
		$dpar = "";
		if ($dp) {
			if (($dp =~ m/^(\d+)/) && $loaded_rnam{"DPos_$1"}) {
				$d2 = $1;
				$dp =~ s/^\d+\s*(.*)/$1/;
				$dpar = "<<";
				eval "\%dpdir = \%DPos_$d2;";
				foreach $k (keys(%dpdir)) {
					$dpar .= "/$k $dpdir{$k} ";
				}
				$dpar .= ">>";
			}
			else {
				$dp =~ s:\s*([^\s]+)\s*(.*):$2:; # get "word"
				$dpar = $1;
				$dpar = "" if ($dpar eq "null");
				if ($dpar eq "<<") {
					do {
						$dp =~ s:\s*([^\s]+)\s*(.*):$2:; # get "word"
						$d2 = $1;
						$dpar = $dpar . " $d2";
					} until ($d2 eq ">>");
				};
			}
		};
		$fil .= "$dpar $f filter ";
		# determine data characteristics
		$stream_is_compressed = 1 if ($f eq "/LZWDecode");
		$stream_is_compressed = 1 if ($f eq "/DCTDecode");
		$stream_is_compressed = 1 if ($f eq "/CCITTFaxDecode");
		$stream_is_compressed = 1 if ($f eq "/RunLengthDecode");
		$stream_is_binary = 0 if ($f eq "/ASCII85Decode");
		$stream_is_binary = 0 if ($f eq "/ASCIIHexDecode");
	}
	return $fil;
};

sub gen_stream
{
	local ($ref) = @_;
	local (%stream, $dec, $len);
	
	&pos_ref ($ref);
	%stream = &load_dir;
			# resolve indirect values if needed:
	$stream{"Length"} = &get_val($stream{"Length"}) if $stream{"Length"};
	#$stream{"Length1"} = &get_val($stream{"Length1"}) if $stream{"Length1"};
	#$stream{"Length2"} = &get_val($stream{"Length2"}) if $stream{"Length2"};
	#$stream{"Length3"} = &get_val($stream{"Length3"}) if $stream{"Length3"};
	
	print PS "% stream ref $ref\n";
	$dec = &gen_filterChain ($stream{"Filter"}, $stream{"DecodeParms"});
	if ($stream_is_binary && $avoidBinary) {
		$streamOutMode = $defaultAsciiMode;
		$dec = "$defaultAsciiFilter filter $dec";
	}
	print PS " currentfile $dec cvx exec\n" unless ($dec =~ m/^\s*$/);
	$len = $stream{"Length"};
	# print STDERR "CALL(4): copy_stream_len ($len, PS)\n";
	&copy_stream_len ($len, PS);
	print PS "\n\n%% END stream ref $ref\n";
};

# test if a given font exists on this machine (NEXTSTEP only !!!)
sub font_exists_here
{
	local ($fname) = @_;
	local ($n);
	
	if (-f ($n = "$ENV{'HOME'}/Library/Fonts/$fname.font/$fname") ||
	    -f ($n = "/LocalLibrary/Fonts/$fname.font/$fname") ||
	    -f ($n = "/NextLibrary/Fonts/$fname.font/$fname")) {
		print PS "% $fname exists locally\n";
		return $n;
	};
	return 0;
};

sub loadType3font
{
	local ($bfont, $fref) = @_;
	local (%fdir, $bbox, $fmatrix, $first, $last, $size);
	local ($cpref, %cpdir, %cstream, $glyph, $len, $fil);
	
	return if $loaded{$bfont};
	$loaded{$bfont} = 1;
	
	&pos_ref ($fref);
	%fdir = &load_dir;
	$bbox = $fdir{'FontBBox'};
	$fmatrix = $fdir{'FontMatrix'};
	$cpref = $fdir{'CharProcs'};
	$first = $fdir{'FirstChar'};
	$last = $fdir{'LastChar'};

	&pos_ref ($cpref);
	%cpdir = &load_dir;

	print PS "% make Type3 font $bfont:\n";
	$size = $last - $first + 1 + 1;		# second +1 for /.notdef
	print PS " $bbox $fmatrix $size beginSetupType3font\n";
	print PS " CharProcs begin\n";
	local ($op, $target);
	foreach $glyph (keys(%cpdir)) {
		&pos_ref ($cpdir{$glyph});
		%cstream = &load_dir;
		$fil = &gen_filterChain ($cstream{"Filter"}, $cstream{"DecodeParms"});
		$len = $cstream{"Length"};
		if ($len > 65500) {
			warn "Type3 character description too large (> 64KB), ignored";
			print PS "% /$glyph: description to large (>64KB), make dummy\n";
			print PS " FilProcs /$glyph {} put\n";
			print PS " /$glyph () def\n"; # should at least move by it's width
			# MISSING_CODE: move right by width of character
			next;
		}
		else {
			$op = "copy_to_string";
			$target = "/$glyph";
		}
		print PS " FilProcs /$glyph {$fil} put\n";
		if ($stream_is_binary && $avoidBinary) {
			$streamOutMode = $defaultAsciiMode;
			$op .= $defaultAsciiProcSuffix;
		}
		print PS " $target $len $op\n";
		&copy_stream_len ($len, PS);
	};

	print PS " end $bfont endSetupType3font\n";
};

sub font_prolog
{
	local($fname, $fref) = @_;
	local(%font, %encoding, $ftype, $enc, $basefont, $diffname, $nam);
	local($style, $fdesc);
	
	return if $loaded{$fname};
	$loaded{$fname} = 1;	# auch bei fehler nicht noch mal laden
	
	&pos_ref ($fref);
	%font = &load_dir;
	# &dump_dir (%font);
	$ftype = $font{"Subtype"};
	$nam = $font{"Name"};
	$basefont = $font{'BaseFont'};
	# print "FNT: $fname, $basefont, $ftype\n";
	if (($ftype eq "/Type1") || ($ftype eq "/MMType1")) {
	  if ($basefont =~ m/_/) {
		local ($mmname) = $basefont;
		local ($mmbase, $ff);
		
		$mmname =~ s:^/::;
		$mmname =~ s/,.*$//;
		$mmname =~ s/^[A-Z]{6}\+//;
		$mmbase = $mmname;
		$mmbase =~ s/_[0-9A-Za-z_]+$//;
		# print "$fname is a $ftype font ($mmname / $mmbase)";
		if (($ff = &font_exists_here ($mmbase)) && (! $loaded{$mmbase})) {
			print PS "%  $mmbase included for $fname\n";
			# Font must be in prolog for e.g. MyriadMM_650_300_ names to work
			print PS "%%BeginResource: font $mmbase\n";
			open (MM, "<$ff");
			while (<MM>) {
				print PS;
			};
			close (MM);
			print PS "\n%%EndResource\n";
			$loaded{$mmbase} = 1;
		}
		print PS "%  $mmname: will just be generated from $mmbase\n";
	  }
	}
	elsif ($ftype eq "/TrueType") {
		$basefont = "TrueType_$fname" unless $basefont;
		warn "$fname: Can't deal with $ftype fonts yet (faked !!! ???)\n";
		print PS " /$fname [] /Times-Roman mk_diff_font  % faked !!!???\n";
		return;
	}
	elsif ($ftype eq "/Type3") {
		$basefont = "/Type3_$fname" unless $basefont;
		if ($noType3) {
		print PS " $basefont [] /Times-Roman mk_diff_font  % faked upon request\n";
		}
		else {
			&loadType3font ($basefont, $fref);
		}
	}
	else {
		$basefont = "Unknown_$fname" unless $basefont;
		print PS "%  $fname: Unknown font type $ftype\n";
		print PS "/$fname [] /Times-Roman mk_diff_font %% (faked!)\n";
		warn "$fname: Can't deal with $ftype fonts (faked!)\n";
		# $prolog_error = 1;
		return;
	}

	$style = '';
	($basefont, $style) = split(/,/, $basefont, 2) if ($basefont =~ /.*,.*/);
	$DocFonts{$basefont}++; # Liste aller verwendeten (base-) fonts

	if ( (! $FontBase{$basefont}) && ($ftype ne "/Type3") ) {
			# should include font here !!!! ???
		$fdesc = $font{"FontDescriptor"};
		if (! $fdesc) {
			print "ERROR: FontDescriptor missing for $fname\n";
			print PS "% FontDescriptor missing for $fname\n";
			$prolog_error = 1;
		};
		print PS "%  load font $basefont for $fname here:\n";
		&load_font_desc($fdesc) unless ($ftype eq "/MMType1");
		$FontBase{$basefont} = 100;		# mark as "loaded", nonstandard
	};

	print PS "% load font $fname here (ref $fref, BaseFont = $basefont):\n";
	if (! ($enc = $font{"Encoding"})) {	# if no /Encoding entry
			# take original font encoding, no differences
		print PS "/$fname [] $basefont mk_diff_font\n";
		return;
	};
	if ($enc =~ m:/\w+Encoding:) {
		$enc =~ s/\W*(\w+)\W*/$1/;
		if (   ($enc ne "MacRomanEncoding")
			&& ($enc ne "MacExpertEncoding")
			&& ($enc ne "WinAnsiEncoding")) {
			die "Unknown font encoding $enc\n ";
		};
		if (   ($enc ne "MacRomanEncoding")
			&& ($enc ne "WinAnsiEncoding")) {
			print "WARNING: $enc not yet supported - ";
			$enc = "StandardEncoding";
			print "will use $enc\n";
		};
		# mit vordefinierten Encoding Vektoren
		print PS "/$fname $enc $basefont mk_enc_font\n";
		return;
	};
	# ????????? BaseEncoding , MacRomanEncoding etc. noch behandeln !!!
	if ($enc =~ /(\d+)\s+(\d+)\s+R/) {
		$diffname = "EncObj_$1_$2_Diff";
		if ( ! $loaded{$diffname} ) {
			print PS "% encoding from object $enc\n";
			&pos_ref ($enc);
			%encoding = &load_dir;
			$enc = $encoding{"Differences"};
			print PS "/$diffname $enc def\n";
			$loaded{$diffname} = 1;
			# &dump_dir (%encoding);
		};
	}
	else {
		$diffname = $fname . "_Diffs";
		# ??? hier noch mehrfach def's von /Diffs vermeiden
		print PS "/$diffname $enc def\n";
	}
	print PS "/$fname $diffname $basefont mk_diff_font\n";
};

sub load_font_desc
{
	local($fd) = @_;
	local(%fddir, $ffile, $fname);
	
	&pos_ref ($fd);
	%fddir = &load_dir;
	die "Error in FontDescriptor\n" if ($fddir{"Type"} ne "/FontDescriptor");
	$ffile = $fddir{"FontFile"};
	$fname = $fddir{"FontName"};
	$fname =~ s:^/::;
	if ($ffile) {	# embedded Type1 Font
		print PS "%%BeginResource: font $fname\n";
		&gen_stream ($ffile);
		print PS "\n%%EndResource\n";
	}
	elsif (! &font_exists_here ($fname)) {
		print PS "% Can't yet replace font $fname\n";
		print PS "/$fname [] /Times-Roman mk_diff_font % (faked!)\n";
		print "% Can't yet replace font $fname (use Times-Roman instead)\n";
	}
	else {
		print PS "%% $fname is an existing font - wait and see\n";
	};
};

sub get_val
{
	local($val) = @_;
	if ($val =~ /^\s*(\d+)\s+(\d+)\s+R\s*$/) {
		local($tmppos) = tell (PDF);
		&pos_ref ($val);
		$val = &token;
		seek (PDF, $tmppos, 0);
	}
	return $val;
}

sub get_ref_array
{
	local($ref) = @_;
	local($val, $tok, $array_nest, $tmppos);
	
	if ($ref =~ /^\s*(\d+)\s+(\d+)\s+R\s*$/) {
		$tmppos = tell (PDF);
		&pos_ref ($ref);
		$tok = &token;
		if ($tok ne "[") {
			die "get_ref_array: array expected, found '$tok' !!";
		};
		if ($tok eq "[") {
			$array_nest = 1;	# first '[' already scanned
			$val = $tok;
			do {
				$tok = &token;
				$val .= " ".$tok;
				$array_nest++ if ($tok eq "[");
				$array_nest-- if ($tok eq "]");
			} until (($array_nest == 0) && ($tok eq "]"));
		}
		seek (PDF, $tmppos, 0);
	}
	else {
		die "get_ref_array: '$ref' must be a reference !!";
	};
	return $val;
}

sub enc85
{
	local ($val) = @_;
	local ($tmp, $b1, $b2, $b3, $b4, $b5);
	use integer;
	
	# print "TRACE: enc85: val=$val\n";
	return "z" if ($val == 0);
	
    if ($val < 0) {	# Because some don't support unsigned long 
      $tmp = 32;
      $val = $val - (85 * 85 * 85 * 85 * 32);
    }
    if ($val < 0) {
      $tmp = 64;
      $val = $val - (85 * 85 * 85 * 85 * 32);
    }
    $b1 = (($val / (85 * 85 * 85 * 85)) + $tmp) + 33;
    $val %= (85 * 85 * 85 * 85);
    $b2 = ($val / (85 * 85 * 85)) + 33;
    $val %= (85 * 85 * 85);
    $b3 = ($val / (85 * 85)) + 33;
    $val %= (85 * 85);
    $b4 = ($val / 85) + 33;
    $val %= 85;
    $b5 = ($val) + 33;
	return pack ("C5", $b1, $b2, $b3, $b4, $b5);
}

# $blen MUST be a multiple of 4 exept for the last chunk
sub encodeASCII85
{
	local ($blen, $buf, $last, $OUTFILE) = @_;
	local ($i, $v, $off);
	
	$off = 0;
	while ($blen >= 4) {
		$v = unpack ("L", substr ($buf, $off, 4));
		$off += 4;
		$blen -= 4;
		$i = &enc85 ($v);
		$pos85 += length ($i);
		print $OUTFILE $i;
		if ($pos85 > 72) {
			print $OUTFILE "\n";
			$pos85 = 0;
		}
	}
	if ($blen && (! $last)) {
		warn "internal: ascii base-85 encoding called inconsistently";
	}
	if ($blen) {
		$v = 0;
		$i = 0;
		while ($i < $blen) {
			$v = $v * 256 + unpack ("C", substr ($buf, $off, 1));
			$off++;
			$i++;
		}
		while ($i < 4) {
			$v = $v * 256;
			$i++;
		}
		print $OUTFILE substr (&enc85 ($v), 0, $blen + 1);
		$pos85 += $blen + 1;
	}
	if ($last) {
		print $OUTFILE "~>";
		$pos85 = 0;
	}
}

sub copyAndEnc85
{
	local ($encLen, $INFILE, $OUTFILE) = @_;
	local ($val, $i, $outstr);
	use integer;
		
	# print "TRACE: copyAndEnc85: encLen=$encLen\n";
	$outstr = "";
	$val = 0;
	while ($encLen >= 4) {
		$val = ord (getc $INFILE);
		$val = ($val * 256) + ord (getc $INFILE);
		$val = ($val * 256) + ord (getc $INFILE);
		$val = ($val * 256) + ord (getc $INFILE);
		$encLen -= 4;
		$outstr .= &enc85 ($val);
		if (length ($outstr) > 66) {
			print $OUTFILE "$outstr\n";
			$outstr = "";
		}
	}
	if ($encLen) {
		$val = 0;
		$i = 0;
		while ($i < $encLen) {
			$val = $val * 256 + ord (getc $INFILE);
			$i++;
		}
		while ($i < 4) {
			$val = $val * 256;
			$i++;
		}
		$outstr .= substr (&enc85 ($val), 0, $encLen + 1);
	}
	print $OUTFILE "$outstr~>";
}

sub encodeHex
{
	local ($blen, $buf, $last, $OUTFILE) = @_;
	$blen *= 2;
	printf $OUTFILE "%s%s\n", unpack ("H$blen", $buf), ($last) ? ">" : "";
}

sub copy_stream_len
{	# copy stream from PDF current position to $OUTFILE
	local ($cpSlen, $OUTFILE) = @_;
	local ($l, $n, $buf, $chunksize);
	
	$l = $cpSlen;
	$cpSlen = &get_val ($cpSlen); # might be a reference, won't harm otherwise
	# print "TRACE: copy_stream_len: len=$cpSlen (oldl=$l)\n";
	
	while (<PDF>) {
		last if (/^stream/);
	};
	$n = tell(PDF);
	seek (PDF, $n - 6, 0);
	while (1) {
		$c = getc(PDF);
		last if ($c eq "m");	# "streaM"
	}
	$c = getc (PDF); # get LF (or CR) char behind word "stream"
	$n = tell(PDF);
	$c = getc (PDF) if ($c eq "\015"); # get LF if previous was CR
	if ($c ne "\012") {
		warn "Illegal chars after 'stream': (CR) not followed by LF (PDF pos $n)\n"
			unless ($pdf_version == 1000);
		seek (PDF, $n, 0);
	}
	
	$chunksize = 1024;	# for direct copy
	$chunksize = 38 if ($streamOutMode == 1);	# HexEncoded
	# $chunksize = 56 if ($streamOutMode == 2);	# ASCII85Encoded ???
	if ($streamOutMode == 2) {
		if (1) {
			&copyAndEnc85 ($cpSlen, PDF, $OUTFILE);
		}
		else {	local ($tnam) = "/tmp/_e85_$$";
			open (E85, "|encode85>$tnam") || die "PIPE setup";
			while ($cpSlen > 0) {
				$l = $chunksize;
				$l = $cpSlen if ($cpSlen < $chunksize);
				$n = read (PDF, $buf, $l);
				warn "stream read: wanted $l bytes, got only $n" if ($n != $l);
				print E85 $buf;
				$cpSlen -= $n;
			};
			close E85;
			open (I85, "<$tnam");
			while ($n = read (I85, $buf, 1024)) {
				print $OUTFILE $buf;
			}
			close I85;
			print $OUTFILE "~>";
			unlink $tnam;
		}
	}
	else {
	while ($cpSlen > 0) {
		$l = $chunksize;
		$l = $cpSlen if ($cpSlen < $chunksize);
		$n = read (PDF, $buf, $l);
		warn "stream read: wanted $l bytes, got only $n" if ($n != $l);
		if ($streamOutMode == 0) {
			print $OUTFILE "$buf";
		} elsif ($streamOutMode == 1) {
			&encodeHex ($n, $buf, ($n == $cpSlen), $OUTFILE);
		} elsif ($streamOutMode == 2) {
			&encodeASCII85 ($n, $buf, ($n == $cpSlen), $OUTFILE);
		} else {
			warn "internal: illegal streamOutMode ($streamOutMode)\n";
			print $OUTFILE "$buf";
		}
		$cpSlen -= $n;
	};
	}
	$_ = <PDF>;
	$_ = <PDF> if (/^[\012\015]*$/);
	unless (m:^endstream:){ local ($p); $p = tell (PDF);
		warn "stream read: 'endstream' not found (p=$p)" ;
		open (ERR, ">>/tmp/PDF_ERR");
		print ERR "stream read (p=$p): 'endstream' not found in '$_'\n";
	};
	$streamOutMode = 0;
	# while (<PDF>) {
	# 	last if (/^endstream/);
	# 	print $OUTFILE $_;
	# };
};

sub cs_prolog
{
	local($csname, $csref) = @_;
	local(%csdir);
	
	return if $loaded{$csname};
	#$loaded{$csname} = 1;	# auch bei fehler nicht noch mal laden
	
	### ??? not yet finished !!!!
	
	#&pos_ref ($csref);
	#%csdir = &load_dir;
	# &dump_dir (%csdir);
};

sub do_xobj_form
{
	local ($xname, %xdir) = @_;
	local ($ftype, $BBox, $filter, $dp, $matrix);
	
	$ftype = $xdir{"FormType"};
	if ($ftype != 1) {
		print PS "% $xname: Can't deal with $ftype FormType (faked)\n";
		print PS "/$xname { } def\n";
		warn "$xname: Can't deal with $ftype FormType (faked)\n";
		return;
	}
	$len = $xdir{"Length"};
	$len = &get_val($len);
	$BBox = $xdir{"BBox"};
	$matrix = $xdir{"Matrix"};

	print PS "% load XObject $xname here:\n";
	print PS "/$xname { 5 dict\nbegin\n";
	if ($len > 65500) {
		print PS " /.pa_cnt 0 def\n";
		print PS " /_pdf_xobject_$xname {\n";
		print PS "  _pdf_xobjectArray_$xname .pa_cnt get\n";
		print PS "  /.pa_cnt .pa_cnt 1 add def\n";
		print PS " } def\n";
	}
	print PS " << /FormType 1 /BBox $BBox\n   ";
	print PS "/Matrix $matrix\n";

	$dp = $xdir{"DecodeParms"};
	print PS "% DecodeParms = $dp\n" if ($dp);
	$filter = &gen_filterChain ($xdir{"Filter"}, $dp);
	print PS "   /PaintProc { pop _pdf_xobject_$xname \n";
	print PS "    $filter cvx exec }\n"; 

	print PS " >> execform\nend } def\n";

	print PS "% store /$xname data:\n";
	local ($op, $target);
	if ($len > 65500) {
		$op = "copy_to_procarray";
		$target = "/_pdf_xobjectArray_$xname";
	}
	else {
		$op = "copy_to_string";
		$target = "/_pdf_xobject_$xname";
	}
	if ($stream_is_binary && $avoidBinary) {
		$streamOutMode = $defaultAsciiMode;
		$op .= $defaultAsciiProcSuffix;
	}
	print PS " $target $len $op\n";
	
	&copy_stream_len ($len, PS);
	print PS "\n%%_pdf_End_of_Data\n";
}

sub xobj_prolog
{
	local($xname, $xref) = @_;
	local(%xdir, $xtype, $nam, $w, $h, $bpc, $cs, $hcs, $filter, $tmppos);
	local($len, $imask, $dp);
	
	return if $loaded{$xname};
	$loaded{$xname} = 1;	# auch bei fehler nicht noch mal laden
	
	&pos_ref ($xref);
	%xdir = &load_dir;
	# &dump_dir (%xdir);
	$xtype = $xdir{"Subtype"};
	$nam = $xdir{"Name"};
	if ($xtype ne "/Image") {
		# $prolog_error = 1;
		if ($xtype eq "/Form") {
			&do_xobj_form ($xname, %xdir);
			return;
		};
		print PS "% $xname: Can't deal with $xtype XObjects yet (faked)\n";
		print PS "/$xname { } def\n";
		warn "$xname: Can't deal with $xtype XObjects yet (faked)\n";
		return;
	};
	$w = $xdir{"Width"};
	$h = $xdir{"Height"};
	$bpc = $xdir{"BitsPerComponent"};
	$imask = ($xdir{"ImageMask"} eq "true");
	$len = $xdir{"Length"};
	$len = &get_val($len);
	$cs = $xdir{"ColorSpace"};
	$dec = $xdir{"Decode"};
	$hcs = $cs;
	$hcs = &get_val($cs) if ($cs =~ /^\d+\s+\d+\s+R$/);
	if ($hcs eq "[") {
		$hcs = &get_ref_array($cs) if ($cs =~ /\d+\s+\d+\s+R/);
		$cs = "/Indexed";
		print "/Indexed Image on page $pg\n";
	}
	else {
		$cs = $hcs;
	};
	$dec = $Decode{$cs} unless ($xdir{"Decode"});
	if ($hcs =~
	  m:\[\s*/Indexed\s*(/Device[A-Z]+)\s+(\d+)\s+(\d+\s+\d+\s+R)\s*\]:) {
	  	local ($base, $hival, $lookup, $lnam);
		  	$base = $1;
			$hival = $2;
			$lookup = $3;
			$lnam = $lookup;
			$lnam =~ s: R::;
			$lnam =~ s: :_:g;
			$lnam = "inx_lookup_obj_$lnam";
			&load_print_index_lookup($lookup, $lnam, PS)
				unless $loaded{$lnam};
			$loaded{$lnam} = 1;
			$hcs = "[/Indexed$base $hival $lnam]";
			$dec = "[0 $hival]" unless ($xdir{"Decode"});
	};
	
	print PS "% load XObject $xname here:\n";
	print PS "/$xname { 5 dict\nbegin\n";
	if ($len > 65500) {
		print PS " /.pa_cnt 0 def\n";
		print PS " /_pdf_xobject_$xname {\n";
		print PS "  _pdf_xobjectArray_$xname .pa_cnt get\n";
		print PS "  /.pa_cnt .pa_cnt 1 add def\n";
		print PS " } def\n";
	}
	print PS " $hcs setcolorspace\n" unless $imask;
	# print PS " gsave [$w 0 0 -$h 0 $h] concat mk_cross grestore\n";
	print PS " gsave mk_cross grestore\n" unless ($imask); # shines through
	print PS " << /ImageType 1 /Width $w /Height $h\n   ";
	print PS "/ImageMatrix [$w 0 0 -$h 0 $h]\n";

	$dp = $xdir{"DecodeParms"};
	print PS "% DecodeParms = $dp\n" if ($dp);
	$filter = &gen_filterChain ($xdir{"Filter"}, $dp);
	print PS "   /DataSource { _pdf_xobject_$xname }\n";
	print PS "    $filter\n"; 

	print PS "   /BitsPerComponent $bpc";
	print PS " /Decode $dec" if $dec;
	if ($imask)
		{ print PS " /Decode [0 1] " unless $dec;
		  print PS " >> imagemask\nend } def\n";
		}
	else
		{ print PS " >> image\nend } def\n";}
	if (0 && ($xdir{"Filter"} =~ m:/CCITTFaxDecode:)) {
	# bug in NeXT DiplayPostScript interpreter up to NEXTSTEP version 3.2
	# fixed since NEXTSTEP 3.3 (delete the "0 &&" above if you run <= 3.2)
		print PS "% image $xname omitted due to bug in PS CCITTFaxDecode\n";
		print "% image $xname omitted due to bug in PS CCITTFaxDecode\n";
		print PS "/$xname { gsave mk_cross /Helvetica findfont 0.05 scalefont
			setfont 0.2 0.4 moveto ($xname) show grestore } def
			% override def to avoid any harm\n";
	}
	print PS "% store /$xname data:\n";
	local ($op, $target);
	if ($len > 65500) {
		$op = "copy_to_procarray";
		$target = "/_pdf_xobjectArray_$xname";
	}
	else {
		$op = "copy_to_string";
		$target = "/_pdf_xobject_$xname";
	}
	if ($stream_is_binary && $avoidBinary) {
		$streamOutMode = $defaultAsciiMode;
		$op .= $defaultAsciiProcSuffix;
	}
	print PS " $target $len $op\n";
	
	# print STDERR "CALL(1): copy_stream_len ($len, PS)\n";
	&copy_stream_len ($len, PS);
	print PS "\n%%_pdf_End_of_Data\n";
};

sub scan_outlines
{
	local ($ort) = @_;
	local (%odir, $l, $f);
	
	return unless $ort;
	%odir = &load_dir_by_ref ($ort);
	$l = $odir{"Last"};
	$f = $odir{"First"};
	return unless $f;
	&scan_one_level ($f, $l, ":= ");
}

sub scan_one_level
{
	local	($el, $end, $pref) = @_;
	local	(%edir, $i);
	
	# print "$pref--- Start (sub-)list\n";
	$i = 0;
	while (1) {
		$i++;
		%edir = &load_dir_by_ref ($el);
		print "$pref $i. $edir{'Title'}   ==> $edir{'Dest'}\n";
		if ($edir{"Count"}) {
			&scan_one_level ($edir{"First"}, $edir{"Last"}, "$pref  "); 
		};
		last if ($el eq $end);
		$el = $edir{"Next"};
	}
	# print "$pref--- End list ($i entries)\n";
	return;
}

sub init
{		# Base-14 fonts (always defined in PDF)
	$FontBase{"/Courier"} = 14;
	$FontBase{"/Courier-Bold"} = 14;
	$FontBase{"/Courier-Oblique"} = 14;
	$FontBase{"/Courier-BoldOblique"} = 14;
	$FontBase{"/Helvetica"} = 14;
	$FontBase{"/Helvetica-Bold"} = 14;
	$FontBase{"/Helvetica-Oblique"} = 14;
	$FontBase{"/Helvetica-BoldOblique"} = 14;
	$FontBase{"/Times-Roman"} = 14;
	$FontBase{"/Times-Bold"} = 14;
	$FontBase{"/Times-Italic"} = 14;
	$FontBase{"/Times-BoldItalic"} = 14;
	$FontBase{"/Symbol"} = 14;
	$FontBase{"/ZapfDibgbats"} = 14;
		# Base-35 fonts (found in most PostScript printers)
	$FontBase{"/AvantGarde-Book"} = 35;
	$FontBase{"/AvantGarde-BookOblique"} = 35;
	$FontBase{"/AvantGarde-Demi"} = 35;
	$FontBase{"/AvantGarde-DemiOblique"} = 35;
	$FontBase{"/Bookman-Demi"} = 35;
	$FontBase{"/Bookman-DemiItalic"} = 35;
	$FontBase{"/Bookman-Light"} = 35;
	$FontBase{"/Bookman-LightItalic"} = 35;
	$FontBase{"/Helvetica-Narrow"} = 35;
	$FontBase{"/Helvetica-Narrow-Bold"} = 35;
	$FontBase{"/Helvetica-Narrow-Oblique"} = 35;
	$FontBase{"/Helvetica-Narrow-BoldOblique"} = 35;
	$FontBase{"/NewCenturySchlbk-Bold"} = 35;
	$FontBase{"/NewCenturySchlbk-BoldItalic"} = 35;
	$FontBase{"/NewCenturySchlbk-Italic"} = 35;
	$FontBase{"/NewCenturySchlbk-Roman"} = 35;
	$FontBase{"/Palatino-Bold"} = 35;
	$FontBase{"/Palatino-BoldItalic"} = 35;
	$FontBase{"/Palatino-Italic"} = 35;
	$FontBase{"/Palatino-Roman"} = 35;
	$FontBase{"/ZapfDingbats"} = 35;
	
	$Decode{"/DeviceGray"} = "[0 1]";
	$Decode{"/DeviceRGB"} = "[0 1 0 1 0 1]";
	$Decode{"/DeviceCMYK"} = "[0 1 0 1 0 1 0 1]";
	$Decode{"/CIEBasedABC"} = "[0 1 0 1 0 1]";
	$Decode{"/CIEBasedA"} = "[0 1]";
	$Decode{"/Seperation"} = "[0 1]";
	$Decode{"/Indexed"} = "[0 255]";	# Dummy, 255 not ok !!! ???
};
	######### Main Program ############

	print  "Analyzing PDF file '$pdffile'\n";

# load cross refrence table

	&init;
	&find_startxref;
	die "PDF file structure corrupt - can't continue.\n" unless $startxref;

	open (LOG, ">$basename.log") if $logging;

	printf "Xref starts at %d\n", $startxref if $verbose;
	$pos = $startxref;
	while (1) {
		&load_xref ($pos);
		if (1) {
			(/^trailer/) || die "Trailer keyword not recognized\n";
			$tbuf = <PDF>;
		}
		else {
			(&token eq "trailer") || die "Trailer keyword not recognized\n";
			# $tbuf = <PDF>;
		}
		%tmp = &load_dir;
		%document = %tmp unless $document{"Root"}; # preserve "latest" trailer
		$pos = $tmp{"Prev"};
		last unless $pos;
	};

	printf "Root is '%s', Info is '%s'\n",
		$document{"Root"}, $document{"Info"} if $verbose;
	&dump_dir (%document) if $verbose;
	
	if ($document{"Encrypt"}) {
		unlink ("$basename.ps");
		die "\nThis document is encrypted. The en/decryption algorithm is
not documented by Adobe, so I can't convert it - sorry\n\n"
	}
# get catalog object
	print LOG "== get catalog object\n" if $logging;
	&pos_ref ($document{"Root"});
	%catalog = &load_dir;
	if ($verbose) {
		print "Catalog:\n";
		&dump_dir (%catalog);
	};
	# print "Outlines ignored\n" if $catalog{"Outlines"};
	&scan_outlines ($catalog{"Outlines"}) if $verbose;

# get list of page objects

	print LOG "== get list of page objects\n" if $logging;
	$page_cnt = 0;
	&load_pages ($catalog{"Pages"});
	
# write header

	open (PS, ">$basename.ps");
	print "Output File: $basename.ps\n";
	
	print PS "%!PS-Adobe-2.0\n";
	print PS "%%Title: $pdffile\n";
	$Date = &ctime(time);
	$tmp = $0;
	$tmp =~ s:.*/::;	# strip leading path if present
	print PS "%%CreationDate: $Date";
	print PS "%%Creator: $tmp (Version $Version [Rev $Rev], (C) 1993-1995 by D. Droege)\n";
	print PS "%%BoundingBox: 0 0 612 842\n";
	print PS "%%Pages: $page_cnt 1\n";
	print PS "%%For: $ENV{'USER'}@$ENV{'HOST'}\n";
	print PS "%%DocumentNeededResources: (atend)\n";
	print PS "%%DocumentSuppliedResources: (atend)\n";# ??? avoid if not needed
	print PS "%%LanguageLevel: 2\n";	# needed for 'filter' etc.
	print PS "%%EndComments\n";
	
	print PS "\n%%BeginProlog\n\n";
	print PS "systemdict /languagelevel known\n";
	print PS " {languagelevel 2 ge}{false}ifelse not\n";
	print PS " {(== Must run on PostScript Level 2 device ==)== stop} if\n\n";
	print PS "\n%%BeginResource: procset PDF_ProcSet\n 1 setobjectformat\n";
	&put_proc_set;
	print PS " /copy_to_string_Hex { % /strname length copy_to_string_Hex
        string currentfile /ASCIIHexDecode filter exch readstring pop def
  } bd\n";
	print PS " /copy_to_string_A85 { % /strname length copy_to_string_A85
        string currentfile /ASCII85Decode filter exch readstring pop def
  } bd\n";
	print PS " /copy_to_procarray_fil { % /paname length file copy_to_procarray
        /.xfile exch def
        /.xlen exch def
        /.xnamA exch def
        [
         {
          .xlen 0 le { exit } if
          .xlen 4096 le { .xlen } { 4096 } ifelse
          dup .xlen exch sub /.xlen exch def
          string .xfile exch readstring pop
         } loop
         ()     % empty string as EOD to signal end of data
        ] .xnamA exch def
  } bd\n";
	print PS " /copy_to_procarray_Hex { % /paname length copy_to_procarray_Hex
        currentfile /ASCIIHexDecode filter copy_to_procarray_fil
  } bd\n";
	print PS " /copy_to_procarray_A85 { % /paname length copy_to_procarray_A85
        currentfile /ASCII85Decode filter copy_to_procarray_fil
  } bd\n";

	print PS "\n%%EndResource\n\n";
	
	print LOG "== collect needed resources/writing prolog\n" if $logging;
	print "Writing out $page_cnt page resources ...\n";
	for ($pg = 1; $pg <= $page_cnt; $pg++) {
		&gen_page_resource ($pg);
		# print STDERR "$pg " if (($pg % 10) == 0);
	};
	print " generated $XObjects XObjects\n" if $XObjects;
	print " $cnt_annots Annotations ignored\n" if ($verbose && $cnt_annots);
	print PS "\n%%EndProlog\n";
	if ((! $NoThumbs) && $setup_thumb_done) {
		print THUMB "showpage\n%%Trailer\n%%Pages: $thumb_pg 1\n";
		close THUMB;
	};
	# print STDERR "\n";
	
	die "Error while generating PostScript Prolog\n" if $prolog_error;
	
	print LOG "== generate pages\n" if $logging;
	for ($pg = 1; $pg <= $page_cnt; $pg++) {
		print STDERR "[$pg";
		print PS "\n%%Page: $pg $pg\n";
		&gen_page_output ($pg);
		print STDERR "] ";
		print STDERR "\n" if (($pg % 10) == 0);
	};
	print STDERR "\n" if (! ((($pg - 1) % 10) == 0));
	# print PS "usertime startTime sub 1000 div 20 string cvs ==\n";
	print PS "%%Trailer\n";
		# ??? veraltet, umstellen auf ...SuppliedResurces ...
	$first = 1;
	foreach $fnt (sort keys(%DocFonts)) {
		next if ($FontBase{$fnt} > 35);
		$fnt =~ s:^/::;
		if ($first == 1) {
			print PS "%%DocumentNeededResources: font $fnt\n";
			$first = 0;
		}
		else {
			print PS "%%+ font $fnt\n";
		};
	};
	foreach $fnt (sort keys(%DocFonts)) {
		next if ($FontBase{$fnt} > 35);
		$fnt =~ s:^/::;
		print PS "%%IncludeResource: font $fnt\n";
	};
	print PS "%%DocumentSuppliedResources: procset PDF_ProcSet\n";
	foreach $fnt (sort keys(%DocFonts)) {
		next if ($FontBase{$fnt} <= 35);
		$fnt =~ s:^/::;
		print PS "%%+ font $fnt\n";
	};
	print PS "%%EOF\n";
